/*
    This file is part of jLensInfo by Philipp Hoehne <http://digicam4.me> 

    jLensInfo  is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    jLensInfo is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jLensInfo.  If not, see <http://www.gnu.org/licenses/>.
 */
   
package com.phoehne.exif.lensinfo.canon;

import java.io.InputStream;
import java.util.HashMap;
import java.util.Scanner;

/**
 *
 * @author dorian
 */
public class CanonLensNameExtractor {

    private static HashMap lensNameMap;
    private static HashMap reverseLookupMap;
    private HashMap roundingMap;
    private final static String UNKOWN_LENS = "Unknown lens on Canon";

    public CanonLensNameExtractor() {
        // @todo Throw exception when initialization failes
        // Init lens names list
        lensNameMap = new HashMap();
        reverseLookupMap = new HashMap();
        roundingMap = new HashMap();
        getMap(lensNameMap,"CanonLenses");
        getMap(reverseLookupMap,"CanonLensesReverseLookup");
        
        // converting apertureMin and apertureMax sometimes produces wrong results due to rounding
        // e.g.: Canon EOS REBEL T4i with EF-S18-55mm f/3.5-5.6 IS : 3.550752985053042 => 3.6 
        // I wasn't able to figure out how exiftool and exiv2 solve this problem :-(
        roundingMap.put("0.9","1"); // very unlikely, see "1.1"
        roundingMap.put("1","1");
        roundingMap.put("1.0","1");
        roundingMap.put("1.1","1.2"); // aperture (min/max) tag source value of 32 == f/1.0 doesn't cause fractions so I decided to use 1.1 for 1.2
        roundingMap.put("1.2","1.2");
        roundingMap.put("1.3","1.4");
        roundingMap.put("1.4","1.4");
        roundingMap.put("1.5","1.4");
        roundingMap.put("1.7","1.8");
        roundingMap.put("1.8","1.8");
        roundingMap.put("1.9","1.8"); // aperture (min/max) tag source value of 64 == f/2.0 doesn't cause fractions so I decided to use 1.9 for 1.8
        roundingMap.put("2","2");
        roundingMap.put("2.0","2");
        roundingMap.put("2.1","2");
        roundingMap.put("2.4","2.5");
        roundingMap.put("2.5","2.5");
        roundingMap.put("2.6","2.5");
        roundingMap.put("2.7","2.8");
        roundingMap.put("2.8","2.8");
        roundingMap.put("2.9","2.8");
        roundingMap.put("3.4","3.5");
        roundingMap.put("3.5","3.5");
        roundingMap.put("3.6","3.5");
        roundingMap.put("3.9","4"); // very unlikely, see "1.1"
        roundingMap.put("4","4");
        roundingMap.put("4.0","4");
        roundingMap.put("4.1","4"); // very unlikely, see "1.1"
        roundingMap.put("4.4","4.5");
        roundingMap.put("4.5","4.5");
        roundingMap.put("4.6","4.5");
        roundingMap.put("5.5","5.6");
        roundingMap.put("5.6","5.6");
        roundingMap.put("5.7","5.6");
        roundingMap.put("6.2","6.3");
        roundingMap.put("6.3","6.3");
        roundingMap.put("6.4","6.3");
    }
          

    public synchronized String getName(String _lensModel, long _lensType, int _focalShort, int _focalLong, int _focalUnits, long _apertureMin, long _apertureMax) {

        int focalShort = _focalShort;
        int focalLong = _focalLong;
        int focalUnits = _focalUnits;
        int lensType = (int)_lensType; // Cut of fractions, some lenses come with decimal values, we'll take care of this later @todo Investigate on non-int Canon lensType IDs
        float extendedLensType;
        float apertureMin = _apertureMin;
        float apertureMax = _apertureMax;
        String lensModel = _lensModel.trim();
        String lookupKey;
        Object tempAperture;
        Object tempLens;
        Object tempLensType;
        
        // Correct invalid focal length and aperture (0xffff)
        // Actually somewhat an overhead because of the following logic
        // but hopefully easier to read 
        if(focalShort == 0xffff) focalShort = 0;
        if(focalLong == 0xffff) focalLong = 0;
        if(apertureMin == 0xffff) apertureMin = 0;
        if(apertureMax == 0xffff) apertureMax = 0;
        
        // If there is no lens type but the lens model field contains an entry starting with "EF"
        // then it's a Canon lens (at least from what I've seen so far)
        if(lensType == 0 && lensModel.startsWith("EF")) {
            return "Canon " + lensModel;
        }
        
        // No lense type and nothing in lense model?
        // => Give up 
        if(lensType == 0)
            lensModel = UNKOWN_LENS;
        else {
        
            // Lookup lens, attempt 1
            // Straight-foward lookup by lens type

            tempLens = lensNameMap.get(String.valueOf(lensType));
            if(tempLens != null) { 
                lensModel = tempLens.toString();
            }
            else { // Nothing found? Try the more complex reverse-lookup
            
                // Correct focal length in case focal units != 1 (default) and focal units != 0 (to avoid division by zero errors)
                // We ignore _focalShort == 0 OR _focalLong == 0 here because it causes no harm, result will be simply zero
                if (focalUnits != 1 && focalUnits != 0) {
                    focalShort = focalShort / focalUnits;
                    focalLong = focalLong / focalUnits;
                } 

                // Convert and round aperture values
                //apertureMin = (float)Math.round((Math.exp(canonEv((long)apertureMin) * Math.log(2)/2))*  10)/10; // For future use
                apertureMax = (float)Math.round((Math.exp(canonEv((long)apertureMax) * Math.log(2)/2))*  10)/10;

                /*
                 * Try to fix rounding problem
                 * We won't catch parsing exception as we have full control over the values in the lookup map
                 * and of course never ever will make any mistake there
                 * 
                */

                /* For future use => enable apertureMin calculation above as well!!!
                tempAperture = roundingMap.get(Float.toString(apertureMin));
                if(tempAperture != null) apertureMin = Float.parseFloat(tempAperture.toString());
                */

                tempAperture = roundingMap.get(Float.toString(apertureMax));
                if(tempAperture != null) apertureMax = Float.parseFloat(tempAperture.toString());
                
                // Build the lookup key
                if(focalShort == focalLong || focalLong == 0 ) { // fix focal lense
                    lookupKey = lensType + "-" + apertureMax + "-" + focalShort;        
                }
                else { // Zoom
                    lookupKey = lensType + "-" + apertureMax + "-" + focalShort + "-" + focalLong;    
                }
                
                // Do the reverse-lookup to get the extended lens type id (e.g. 10.0, 10.1 ...)
                tempLensType = reverseLookupMap.get(lookupKey);
                if(tempLensType != null) {
                    extendedLensType = Float.parseFloat(tempLensType.toString());
                    
                    // Now lookup the lens list with the new lookup key
                    tempLens = lensNameMap.get(String.valueOf(extendedLensType));
                    if(tempLens != null) { 
                        lensModel = tempLens.toString();
                    }
                
            }


// For debugging
                System.out.print(
                        focalShort +
                        "-" + focalLong +
                        "-" + apertureMin +
                        "-" + apertureMax + " : "
                        );
            }
        }
        return lensModel;
    }
      
   
    // Convert Canon-style values to f-stop values
    
    private static float canonEv(long val)
    {
        // temporarily remove sign
        int sign = 1;
        if (val < 0) {
            sign = -1;
            val = -val;
        }
        // remove fraction
        float frac = (float)(val & 0x1f);
        val = val - (long)frac; // remove fraction
        // convert 1/3 (0x0c) and 2/3 (0x14) codes
        if (frac == 0x0c) {
            frac = 32.0f / 3;
        }
        else if (frac == 0x14) {
            frac = 64.0f / 3;
        }
        return (float)(sign * (val + frac) / 32.0f);
    }
    
        // Load a hashmap from file with name-value pairs, delimtered by "|" (pipe)
    
        private static void getMap(HashMap mapName, String fileName) {
        InputStream IS;
        Scanner in;
        String[] items;
        IS = CanonLensNameExtractor.class.getResourceAsStream(fileName);
        in = new Scanner(IS);
        while(in.hasNext()) {
            items = in.nextLine().split("\\|");
            mapName.put(items[0], items[1]);
        }
        in.close();
    }
    
}
