/*
    jLensInfo - Example for Canon

    This file is part of jLensInfo by  
    @author Philipp Hoehne <http://digicam4.me> 

    Usage: 
        java -cp jLensInfo.jar com.phoehne.exif.lensinfo.nikon.Example <image>

        e.g.:
            Windows
            java -cp jLensInfo.jar com.phoehne.exif.lensinfo.canon.Example samples\nikon.jpg

            OSX/ Linux:
            java -cp jLensInfo.jar com.phoehne.exif.lensinfo.canon.Example /samples/nikon.jpg

    License:

    jLensInfo  is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    jLensInfo is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jLensInfo.  If not, see <http://www.gnu.org/licenses/>.
*/


package com.phoehne.exif.lensinfo.canon;

import com.drew.imaging.ImageMetadataReader;
import com.drew.imaging.ImageProcessingException;
import com.drew.metadata.Directory;
import com.drew.metadata.Metadata;
import com.drew.metadata.exif.ExifIFD0Directory;
import com.drew.metadata.exif.CanonMakernoteDirectory;
import java.io.File;

public class Example {

    public static void main(String[] args) {
        
        Directory ExifIFD0Dir = new ExifIFD0Directory();
        Directory CanonDir = new CanonMakernoteDirectory();
        Directory workDir;
        Metadata metadata;
        File f = null;
        String maker = "Unknown maker";
        String model = "Unknown model";
        String lensModel = "Unknown lens on Canon";
        long lensType = 0;
        int focalShort = 0;
        int focalLong = 0;
        int focalUnits = 1; // assume Units = 1 if not present
        long apertureMin = 0;
        long apertureMax = 0;
        CanonLensNameExtractor myCanonLensNameExtractor;
        
        if(args.length == 1) { 
            f = new File(args[0]);
        }
        else {
            System.out.println("Wrong number of parameters!\nUsage: java -cp jLensInfo.jar com.phoehne.exif.lensinfo.canon.Example <image file> ");
            System.exit(-1);
        }
        
        try {
            
                metadata = ImageMetadataReader.readMetadata(f);
                
                myCanonLensNameExtractor = new CanonLensNameExtractor();

                // Get Maker and Model
                if (metadata.containsDirectory(ExifIFD0Dir.getClass())) {
                    workDir = metadata.getDirectory(ExifIFD0Dir.getClass());
                    if(workDir.containsTag(ExifIFD0Directory.TAG_MAKE)) {
                        maker = workDir.getString(ExifIFD0Directory.TAG_MAKE);
                        if(!maker.toUpperCase().contains("CANON")) {
                            System.out.println("The metadata info in the provided file is corrupt or not a Canon file\nCamera Maker found was: " + maker);
                            System.exit(-1);
                        }
                    }    
                    // Actually the model is not required to identify the lens, so this can be left out
                    if(workDir.containsTag(ExifIFD0Directory.TAG_MODEL)) {
                        model = workDir.getString(ExifIFD0Directory.TAG_MODEL);
                    } 
                    else {  System.out.println("Not able to get the Camera Model name"); 
                            System.exit(-1);                    
                    }
                }
                
                if (metadata.containsDirectory(CanonDir.getClass())) {
                    workDir = metadata.getDirectory(CanonDir.getClass());
                    if(workDir.containsTag(CanonMakernoteDirectory.TAG_LENS_MODEL)) {
                        lensModel = workDir.getString(CanonMakernoteDirectory.TAG_LENS_MODEL);
                    }   
                    // Get Canon lens type (TAG_UNKNOWN_8) 
                    if(workDir.containsTag(CanonMakernoteDirectory.CameraSettings.TAG_UNKNOWN_8 )) {
                        if(workDir.getInt(CanonMakernoteDirectory.CameraSettings.TAG_UNKNOWN_8) != 65535)
                        lensType = workDir.getLong(CanonMakernoteDirectory.CameraSettings.TAG_UNKNOWN_8);
                    }  
                    
                    // Get focal short (in focal units)
                    if(workDir.containsTag(CanonMakernoteDirectory.CameraSettings.TAG_SHORT_FOCAL_LENGTH )) {
                        focalShort = workDir.getInt(CanonMakernoteDirectory.CameraSettings.TAG_SHORT_FOCAL_LENGTH);
                    }                      
                    // Get focal long (in focal units)
                    if(workDir.containsTag(CanonMakernoteDirectory.CameraSettings.TAG_LONG_FOCAL_LENGTH )) {
                        focalLong = workDir.getInt(CanonMakernoteDirectory.CameraSettings.TAG_LONG_FOCAL_LENGTH);
                    }   
                    // Get focal units (in units / mm)
                    if(workDir.containsTag(CanonMakernoteDirectory.CameraSettings.TAG_FOCAL_UNITS_PER_MM )) {
                        focalUnits = workDir.getInt(CanonMakernoteDirectory.CameraSettings.TAG_FOCAL_UNITS_PER_MM);
                    }                    
                    // Get aperture min (TAG_UNKNOWN_10)
                    if(workDir.containsTag(CanonMakernoteDirectory.CameraSettings.TAG_UNKNOWN_10 )) {
                        apertureMin = workDir.getLong(CanonMakernoteDirectory.CameraSettings.TAG_UNKNOWN_10);
                    }         
                    // Get aperture max (TAG_UNKNOWN_9)
                    if(workDir.containsTag(CanonMakernoteDirectory.CameraSettings.TAG_UNKNOWN_9 )) {
                        apertureMax= workDir.getLong(CanonMakernoteDirectory.CameraSettings.TAG_UNKNOWN_9);
                    }    
                    
                    System.out.println(maker + " : " + model + " : " + myCanonLensNameExtractor.getName(lensModel, lensType, focalShort, focalLong, focalUnits, apertureMin, apertureMax));
                    
                    
                }           
                
        }
        catch (java.io.FileNotFoundException FNFE) {
            System.out.println("An error occured:\nCan't find the file: "+ f.getAbsolutePath());
        }
        
        catch (ImageProcessingException IMPE) {
            System.out.println("Image Processing Exception");
        }
        catch (Exception e) {
            System.out.println("An error occured:\n"+e.getLocalizedMessage());
            e.printStackTrace();  // for debugging @todo Remove .printStackTrace
        }
        
        
    }     
    
}


