/*
    This file is part of jLensInfo by Philipp Hoehne <http://digicam4.me> 

    jLensInfo  is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    jLensInfo is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jLensInfo.  If not, see <http://www.gnu.org/licenses/>.

    Credits
    - Metadata Extractor JAVA library: Drew Noakes <https://drewnoakes.com/code/exif/>
    - Nikon Unique Lens ID: Robert Rottmerhusen <http://www.rottmerhusen.com>
    - EXIF Tag Information and Decryption of Nikon Lens Data: 
        - exif tool by Phil Harvey <http://www.sno.phy.queensu.ca/~phil/exiftool/>
        - exiv2 by Andreas Huggel <http://www.exiv2.org>
 */

package com.phoehne.exif.lensinfo.nikon;

import java.util.Formatter;

/**
 *
 * @author dorian
 */
public class NikonLensNameExtractor {
    
    // The lookup table for decryption
    private final static byte  XLAT[][] = {
            { (byte) 0xc1,(byte) 0xbf,(byte) 0x6d,(byte) 0x0d,(byte) 0x59,(byte) 0xc5,(byte) 0x13,(byte) 0x9d,(byte) 0x83,(byte) 0x61,(byte) 0x6b,(byte) 0x4f,(byte) 0xc7,(byte) 0x7f,(byte) 0x3d,(byte) 0x3d,
              (byte) 0x53,(byte) 0x59,(byte) 0xe3,(byte) 0xc7,(byte) 0xe9,(byte) 0x2f,(byte) 0x95,(byte) 0xa7,(byte) 0x95,(byte) 0x1f,(byte) 0xdf,(byte) 0x7f,(byte) 0x2b,(byte) 0x29,(byte) 0xc7,(byte) 0x0d,
              (byte) 0xdf,(byte) 0x07,(byte) 0xef,(byte) 0x71,(byte) 0x89,(byte) 0x3d,(byte) 0x13,(byte) 0x3d,(byte) 0x3b,(byte) 0x13,(byte) 0xfb,(byte) 0x0d,(byte) 0x89,(byte) 0xc1,(byte) 0x65,(byte) 0x1f,
              (byte) 0xb3,(byte) 0x0d,(byte) 0x6b,(byte) 0x29,(byte) 0xe3,(byte) 0xfb,(byte) 0xef,(byte) 0xa3,(byte) 0x6b,(byte) 0x47,(byte) 0x7f,(byte) 0x95,(byte) 0x35,(byte) 0xa7,(byte) 0x47,(byte) 0x4f,
              (byte) 0xc7,(byte) 0xf1,(byte) 0x59,(byte) 0x95,(byte) 0x35,(byte) 0x11,(byte) 0x29,(byte) 0x61,(byte) 0xf1,(byte) 0x3d,(byte) 0xb3,(byte) 0x2b,(byte) 0x0d,(byte) 0x43,(byte) 0x89,(byte) 0xc1,
              (byte) 0x9d,(byte) 0x9d,(byte) 0x89,(byte) 0x65,(byte) 0xf1,(byte) 0xe9,(byte) 0xdf,(byte) 0xbf,(byte) 0x3d,(byte) 0x7f,(byte) 0x53,(byte) 0x97,(byte) 0xe5,(byte) 0xe9,(byte) 0x95,(byte) 0x17,
              (byte) 0x1d,(byte) 0x3d,(byte) 0x8b,(byte) 0xfb,(byte) 0xc7,(byte) 0xe3,(byte) 0x67,(byte) 0xa7,(byte) 0x07,(byte) 0xf1,(byte) 0x71,(byte) 0xa7,(byte) 0x53,(byte) 0xb5,(byte) 0x29,(byte) 0x89,
              (byte) 0xe5,(byte) 0x2b,(byte) 0xa7,(byte) 0x17,(byte) 0x29,(byte) 0xe9,(byte) 0x4f,(byte) 0xc5,(byte) 0x65,(byte) 0x6d,(byte) 0x6b,(byte) 0xef,(byte) 0x0d,(byte) 0x89,(byte) 0x49,(byte) 0x2f,
              (byte) 0xb3,(byte) 0x43,(byte) 0x53,(byte) 0x65,(byte) 0x1d,(byte) 0x49,(byte) 0xa3,(byte) 0x13,(byte) 0x89,(byte) 0x59,(byte) 0xef,(byte) 0x6b,(byte) 0xef,(byte) 0x65,(byte) 0x1d,(byte) 0x0b,
              (byte) 0x59,(byte) 0x13,(byte) 0xe3,(byte) 0x4f,(byte) 0x9d,(byte) 0xb3,(byte) 0x29,(byte) 0x43,(byte) 0x2b,(byte) 0x07,(byte) 0x1d,(byte) 0x95,(byte) 0x59,(byte) 0x59,(byte) 0x47,(byte) 0xfb,
              (byte) 0xe5,(byte) 0xe9,(byte) 0x61,(byte) 0x47,(byte) 0x2f,(byte) 0x35,(byte) 0x7f,(byte) 0x17,(byte) 0x7f,(byte) 0xef,(byte) 0x7f,(byte) 0x95,(byte) 0x95,(byte) 0x71,(byte) 0xd3,(byte) 0xa3,
              (byte) 0x0b,(byte) 0x71,(byte) 0xa3,(byte) 0xad,(byte) 0x0b,(byte) 0x3b,(byte) 0xb5,(byte) 0xfb,(byte) 0xa3,(byte) 0xbf,(byte) 0x4f,(byte) 0x83,(byte) 0x1d,(byte) 0xad,(byte) 0xe9,(byte) 0x2f,
              (byte) 0x71,(byte) 0x65,(byte) 0xa3,(byte) 0xe5,(byte) 0x07,(byte) 0x35,(byte) 0x3d,(byte) 0x0d,(byte) 0xb5,(byte) 0xe9,(byte) 0xe5,(byte) 0x47,(byte) 0x3b,(byte) 0x9d,(byte) 0xef,(byte) 0x35,
              (byte) 0xa3,(byte) 0xbf,(byte) 0xb3,(byte) 0xdf,(byte) 0x53,(byte) 0xd3,(byte) 0x97,(byte) 0x53,(byte) 0x49,(byte) 0x71,(byte) 0x07,(byte) 0x35,(byte) 0x61,(byte) 0x71,(byte) 0x2f,(byte) 0x43,
              (byte) 0x2f,(byte) 0x11,(byte) 0xdf,(byte) 0x17,(byte) 0x97,(byte) 0xfb,(byte) 0x95,(byte) 0x3b,(byte) 0x7f,(byte) 0x6b,(byte) 0xd3,(byte) 0x25,(byte) 0xbf,(byte) 0xad,(byte) 0xc7,(byte) 0xc5,
              (byte) 0xc5,(byte) 0xb5,(byte) 0x8b,(byte) 0xef,(byte) 0x2f,(byte) 0xd3,(byte) 0x07,(byte) 0x6b,(byte) 0x25,(byte) 0x49,(byte) 0x95,(byte) 0x25,(byte) 0x49,(byte) 0x6d,(byte) 0x71,(byte) 0xc7 },
            { (byte) 0xa7,(byte) 0xbc,(byte) 0xc9,(byte) 0xad,(byte) 0x91,(byte) 0xdf,(byte) 0x85,(byte) 0xe5,(byte) 0xd4,(byte) 0x78,(byte) 0xd5,(byte) 0x17,(byte) 0x46,(byte) 0x7c,(byte) 0x29,(byte) 0x4c,
              (byte) 0x4d,(byte) 0x03,(byte) 0xe9,(byte) 0x25,(byte) 0x68,(byte) 0x11,(byte) 0x86,(byte) 0xb3,(byte) 0xbd,(byte) 0xf7,(byte) 0x6f,(byte) 0x61,(byte) 0x22,(byte) 0xa2,(byte) 0x26,(byte) 0x34,
              (byte) 0x2a,(byte) 0xbe,(byte) 0x1e,(byte) 0x46,(byte) 0x14,(byte) 0x68,(byte) 0x9d,(byte) 0x44,(byte) 0x18,(byte) 0xc2,(byte) 0x40,(byte) 0xf4,(byte) 0x7e,(byte) 0x5f,(byte) 0x1b,(byte) 0xad,
              (byte) 0x0b,(byte) 0x94,(byte) 0xb6,(byte) 0x67,(byte) 0xb4,(byte) 0x0b,(byte) 0xe1,(byte) 0xea,(byte) 0x95,(byte) 0x9c,(byte) 0x66,(byte) 0xdc,(byte) 0xe7,(byte) 0x5d,(byte) 0x6c,(byte) 0x05,
              (byte) 0xda,(byte) 0xd5,(byte) 0xdf,(byte) 0x7a,(byte) 0xef,(byte) 0xf6,(byte) 0xdb,(byte) 0x1f,(byte) 0x82,(byte) 0x4c,(byte) 0xc0,(byte) 0x68,(byte) 0x47,(byte) 0xa1,(byte) 0xbd,(byte) 0xee,
              (byte) 0x39,(byte) 0x50,(byte) 0x56,(byte) 0x4a,(byte) 0xdd,(byte) 0xdf,(byte) 0xa5,(byte) 0xf8,(byte) 0xc6,(byte) 0xda,(byte) 0xca,(byte) 0x90,(byte) 0xca,(byte) 0x01,(byte) 0x42,(byte) 0x9d,
              (byte) 0x8b,(byte) 0x0c,(byte) 0x73,(byte) 0x43,(byte) 0x75,(byte) 0x05,(byte) 0x94,(byte) 0xde,(byte) 0x24,(byte) 0xb3,(byte) 0x80,(byte) 0x34,(byte) 0xe5,(byte) 0x2c,(byte) 0xdc,(byte) 0x9b,
              (byte) 0x3f,(byte) 0xca,(byte) 0x33,(byte) 0x45,(byte) 0xd0,(byte) 0xdb,(byte) 0x5f,(byte) 0xf5,(byte) 0x52,(byte) 0xc3,(byte) 0x21,(byte) 0xda,(byte) 0xe2,(byte) 0x22,(byte) 0x72,(byte) 0x6b,
              (byte) 0x3e,(byte) 0xd0,(byte) 0x5b,(byte) 0xa8,(byte) 0x87,(byte) 0x8c,(byte) 0x06,(byte) 0x5d,(byte) 0x0f,(byte) 0xdd,(byte) 0x09,(byte) 0x19,(byte) 0x93,(byte) 0xd0,(byte) 0xb9,(byte) 0xfc,
              (byte) 0x8b,(byte) 0x0f,(byte) 0x84,(byte) 0x60,(byte) 0x33,(byte) 0x1c,(byte) 0x9b,(byte) 0x45,(byte) 0xf1,(byte) 0xf0,(byte) 0xa3,(byte) 0x94,(byte) 0x3a,(byte) 0x12,(byte) 0x77,(byte) 0x33,
              (byte) 0x4d,(byte) 0x44,(byte) 0x78,(byte) 0x28,(byte) 0x3c,(byte) 0x9e,(byte) 0xfd,(byte) 0x65,(byte) 0x57,(byte) 0x16,(byte) 0x94,(byte) 0x6b,(byte) 0xfb,(byte) 0x59,(byte) 0xd0,(byte) 0xc8,
              (byte) 0x22,(byte) 0x36,(byte) 0xdb,(byte) 0xd2,(byte) 0x63,(byte) 0x98,(byte) 0x43,(byte) 0xa1,(byte) 0x04,(byte) 0x87,(byte) 0x86,(byte) 0xf7,(byte) 0xa6,(byte) 0x26,(byte) 0xbb,(byte) 0xd6,
              (byte) 0x59,(byte) 0x4d,(byte) 0xbf,(byte) 0x6a,(byte) 0x2e,(byte) 0xaa,(byte) 0x2b,(byte) 0xef,(byte) 0xe6,(byte) 0x78,(byte) 0xb6,(byte) 0x4e,(byte) 0xe0,(byte) 0x2f,(byte) 0xdc,(byte) 0x7c,
              (byte) 0xbe,(byte) 0x57,(byte) 0x19,(byte) 0x32,(byte) 0x7e,(byte) 0x2a,(byte) 0xd0,(byte) 0xb8,(byte) 0xba,(byte) 0x29,(byte) 0x00,(byte) 0x3c,(byte) 0x52,(byte) 0x7d,(byte) 0xa8,(byte) 0x49,
              (byte) 0x3b,(byte) 0x2d,(byte) 0xeb,(byte) 0x25,(byte) 0x49,(byte) 0xfa,(byte) 0xa3,(byte) 0xaa,(byte) 0x39,(byte) 0xa7,(byte) 0xc5,(byte) 0xa7,(byte) 0x50,(byte) 0x11,(byte) 0x36,(byte) 0xfb,
              (byte) 0xc6,(byte) 0x67,(byte) 0x4a,(byte) 0xf5,(byte) 0xa5,(byte) 0x12,(byte) 0x65,(byte) 0x7e,(byte) 0xb0,(byte) 0xdf,(byte) 0xaf,(byte) 0x4e,(byte) 0xb3,(byte) 0x61,(byte) 0x7f,(byte) 0x2f }
        };  
    
    private byte[] lensData;
    private LensNamesForNikon nikonLensNames;
    private String extendedLensID;
    private byte lensType;
    private int shutterCount;
    private int serialNumber;
    private int lensDataVersion;
    
    public NikonLensNameExtractor() {
        // @todo Throw exception when initialization failes
        // Init lens names list
        nikonLensNames = new LensNamesForNikon();
    }
          
    /* 
     * Extract Lens Data Version from Nikon lensdata
     * The first 4 Bytes of the lens data hold the Nikon Lens ID
     * Attention: there may be some models where this number is not coded as chars!!! Or is this just the case for the Nikon Makernote Version???
     * @todo Identify models where lens ID is not coded as char
     * 
    */

    private int getLensDataVersion() {
        int tmpLensDataVersion = -1; // default, indicates that no valid / known lens data version was found
        char[] tmpChar = new char[4];
        try {
            tmpChar[0] = (char) (0xff & lensData[0]);
            tmpChar[1] = (char) (0xff & lensData[1]);
            tmpChar[2] = (char) (0xff & lensData[2]);
            tmpChar[3] = (char) (0xff & lensData[3]);            
            
            tmpLensDataVersion = Integer.parseInt(new String(tmpChar));         
            
            /* Exceptions may occure when
             * - trying to read the data into the array
             * - converting to an int
             * We don't do anything here, just return the intital value (-1).
             * We may change this in future to have the method throwing an exception
             */
            
        } catch (Exception e) {}
        
        return tmpLensDataVersion;
    }
    
    public synchronized String getName(byte[] _lensData, byte _lensType, int _shutterCount, int _serialNumber) {
        lensData = _lensData;
        lensType = _lensType;
        shutterCount = _shutterCount;
        serialNumber = _serialNumber;
        extendedLensID = "0000000000000000";
        lensDataVersion = getLensDataVersion();
        int startIdx; // This will cause an exception if not correctly initialized later on
        boolean encrypted;  
        boolean hasExtendedLensID; // if false: models which have no extended lens Id but the full lens name coded in the lens info block as Nikon 1 (Lens Types 40x)
        String lensInfoStr = "Unidentified Nikon lens";
        byte[] tempByteArray = new byte[8];


        
        switch (lensDataVersion) {
            case 100    : {
                                encrypted = false;
                                hasExtendedLensID = true;
                                startIdx = 6;
                                break;
            }
            case 101    : {
                                encrypted = false;
                                hasExtendedLensID = true;
                                startIdx = 11;
                                break;
            }
            case 201:
            case 202:
            case 203: {         
                                encrypted = true;
                                hasExtendedLensID = true;
                                startIdx = 11;
                                break;
            }    
            case 204: {         
                                encrypted = true;
                                hasExtendedLensID = true;
                                startIdx = 12;
                                break;
            }      
            /*     
             * Nikon 1's Lens Data has no Lens ID but the full name of the lens encrypted in the 
             * Lens Data Block (64 byte starting from Index 394
             * @todo cleanup and make the array operation fail-save (cases where array is smaller than expected)
             */
            case 400:
            case 401:
            case 402: {         
                                encrypted = true;
                                hasExtendedLensID = false;
                                startIdx = 394;

                                break;
            }     
            default: {         
                                encrypted = true; // Assumption that all future lensdata version will be encrypted
                                hasExtendedLensID = true; // Assumption again
                                startIdx = 12; // Just any value. Lenslookup will most likely fail in this case because the real start index is not known
                                break;
                 }                 
        }
        
       // Step 3: First decrypt the lens data if required
            if(encrypted) {     // Data is encryped
                byte ci, cj, ck;
                int key = 0;
                for (int i = 0; i < 4; ++i) {
                     key ^= (shutterCount >> (i*8)) & 0xff;
                 }
                 ci = XLAT[0][serialNumber & 0xff];
                 cj = XLAT[1][key];
                 ck = 0x60;
                 for (int i = 4; i < lensData.length; ++i) {
                     cj += ci * ck++;
                     lensData[i] ^= cj;
                 }
        }
        
        if(hasExtendedLensID) { // We can construct the extended lens ID

            // Step 4: read LensIdNumber to MCUVersion into temp array
            for (int i=0; i <7; i++) tempByteArray[i] = lensData[startIdx + i];
            //Add Lens ID
            tempByteArray[7] = lensType;

            // Step 5: convert Extebded lens ID into a hex-string
            extendedLensID = bytesToHexString(tempByteArray);
            
            // Step 6: lookup the lens name
            lensInfoStr = nikonLensNames.getLensName(extendedLensID );
        }
        /*     
         * Cameras with lens data version 400.401,402 (e.g.: Nikon 1's) has no Lens ID but the full name of the lens encrypted in the 
         * Lens Data Block (64 byte) starting from Index 394
         * @todo cleanup and make the array operation fail-save (cases where array is smaller than expected)
         */        
        else { 
            byte[] version40xtemp = new byte[64];
            for (int i=0; i < 64; i++) {
                version40xtemp[i] = lensData[startIdx + i];
            }
            lensInfoStr = new String(version40xtemp).trim();  
           
        }
        return lensInfoStr;
    }
      
    private static String bytesToHexString(byte[] bytes) {  
        StringBuilder sb = new StringBuilder(bytes.length * 2);  

        Formatter formatter = new Formatter(sb);  
        for (byte b : bytes) {  
            formatter.format("%02x", b); 
        }  
        return sb.toString().toUpperCase();
    }      
}
